#![feature(proc_macro_hygiene, decl_macro)]

#[macro_use]
extern crate rocket;
#[macro_use]
extern crate rocket_contrib;
#[macro_use]
extern crate diesel;
#[macro_use]
extern crate serde_derive;
extern crate serde_json;

use std::collections::HashMap;

pub mod models;
pub mod schema;
use crate::diesel::GroupedBy;
use crate::models::{Klasse, Stunde, Vertretung};
use crate::schema::klassen;
use crate::schema::stunden;
use crate::schema::vertretungen;
use diesel::prelude::*;
use diesel::BelongingToDsl;
use diesel::QueryDsl;
use diesel::RunQueryDsl;
use rocket::request::{Form, FromParam};
use rocket::response::Redirect;
use rocket_contrib::databases;
use rocket_contrib::serve::StaticFiles;
use rocket_contrib::templates::Template;

struct StundenByID(i32);
#[derive(Debug)]
enum WebError {
    I32NotParsed,
}

impl<'a> FromParam<'_> for StundenByID {
    type Error = WebError;
    fn from_param(param: &rocket::http::RawStr) -> Result<Self, Self::Error> {
        if let Ok(n) = param.parse::<i32>() {
            Ok(StundenByID(n))
        } else {
            Err(WebError::I32NotParsed)
        }
    }
}

#[database("page_db")]
struct DbConn(databases::diesel::SqliteConnection);

#[derive(Debug, Serialize)]
struct ListAllContext {
    klassen: Vec<(Klasse, Vec<Vertretung>)>,
    parent: String,
}

enum QueryClasses {
    All,
    KlassenId(i32),
}

fn query_all_classes(conn: DbConn) -> Vec<(Klasse, Vec<Vertretung>)> {
    let all_classes = klassen::table
        .load::<Klasse>(&*conn)
        .expect("database acess classes");
    let and_substitutions = Vertretung::belonging_to(&all_classes)
        .load::<Vertretung>(&*conn)
        .expect("database acess Vertretungen")
        .grouped_by(&all_classes);
    let combined_classes = all_classes
        .into_iter()
        .zip(and_substitutions)
        .filter(|x| x.1.len() > 0)
        .collect::<Vec<_>>();
    combined_classes
}

fn query_one_class<'a>(
    conn: DbConn,
    id: i32,
) -> (Klasse, HashMap<String, (Stunde, Vec<Vertretung>)>) {
    let mut ret: HashMap<String, (Stunde, Vec<Vertretung>)> = HashMap::new();
    let class = klassen::dsl::klassen
        .filter(klassen::dsl::klassen_id.eq(id))
        .first::<Klasse>(&*conn)
        .expect("database acess classes");
    //let stund: Vec<Stunde> = stunden::table.load(&*conn).expect("Loading Stunden");
    let times: Vec<(Vertretung, Stunde)> = Vertretung::belonging_to(&class)
        .inner_join(stunden::table.on(vertretungen::stunde_id.eq(stunden::stunden_id)))
        .load(&*conn)
        .expect("database access Stunden");
    for (vert, stun) in times {
        let stun_ord = stun.ordinal.to_string();
        let verts = ret.entry(stun_ord).or_insert((stun, Vec::new()));
        verts.1.push(vert);
    }
    (class, ret)
}

#[get("/")]
fn index(conn: DbConn) -> Template {
    let combined_classes = query_all_classes(conn);
    Template::render(
        "uebersicht",
        ListAllContext {
            klassen: combined_classes,
            parent: "base".to_string(),
        },
    )
}

#[derive(Debug, Serialize)]
struct ListOneContext {
    klasse: Klasse,
    vertretungen: HashMap<String, (Stunde, Vec<Vertretung>)>,
    parent: String,
}

#[get("/klasse/<class_id>")]
fn klasse_view(conn: DbConn, class_id: i32) -> Template {
    let class = query_one_class(conn, class_id);
    let (klasse, vertretungen) = class;
    let cont = ListOneContext {
        klasse: klasse,
        vertretungen: vertretungen,
        parent: "base".to_string(),
    };
    Template::render("klasse", cont)
}

#[get("/add/klasse")]
fn get_klasse_form() -> Template {
    Template::render("add_klasse_form", "")
}

#[derive(FromForm, Insertable)]
#[table_name = "klassen"]
struct KlasseForm {
    stufe: i32,
    gruppe: String,
    titel: Option<String>,
    ordnung: i32,
}

#[post("/add/klasse", data = "<klasse>")]
fn post_klasse_form(klasse: Form<KlasseForm>, conn: DbConn) -> Result<Redirect, String> {
    match diesel::insert_into(schema::klassen::table)
        .values(klasse.into_inner())
        .execute(&*conn)
    {
        Ok(_) => Ok(Redirect::to("/admin/add/klasse")),
        Err(_) => Err("Klasse konnte nicht erstellt werden!".to_string()),
    }
}

fn main() {
    rocket::ignite()
        .attach(DbConn::fairing())
        .attach(Template::fairing())
        .mount(
            "/static",
            StaticFiles::from(concat!(env!("CARGO_MANIFEST_DIR"), "/static")),
        )
        .mount("/", routes![index])
        .mount("/", routes![klasse_view])
        .mount("/admin/", routes![get_klasse_form])
        .mount("/admin/", routes![post_klasse_form])
        .launch();
}
